#ifdef __cplusplus
#include <cstdio>
#include <cstdlib>
#include <cstring>
#else
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#endif
#include "ptp_opt.h"
#include "dm.h"

#ifdef PTPDBG_CONFIG_SUPPORT
"  ptpdebug set ptp [enable|disable]\n"
"  ptpdebug set domain <val>\n"
"  ptpdebug set priority1 <val>\n"
"  ptpdebug set priority2 <val>\n"
"  ptpdebug set offset-min-threshold <val>\n"
"  ptpdebug set offset-max-threshold <val>\n"
"  ptpdebug set mpd-threshold <val>\n"
"  ptpdebug set acceptable-master <val>\n"
"  ptpdebug set message-mode [multicast|unicast|mixed]\n"
"  ptpdebug set slave-only [enable|disable]\n"
"  ptpdebug set ttl <val>\n"
"  ptpdebug set dscp-event <val>\n"
"  ptpdebug set dscp-general <val>\n"
"\n"
"  ptpdebug set interface <type> <name> ptp [enable|disable]\n"
"  ptpdebug set interface <type> <name> ipv6 [enable|disable]\n"
"  ptpdebug set interface <type> <name> forced-master [enable|disable]\n"
"  ptpdebug set interface <type> <name> announce-interval <val>\n"
"  ptpdebug set interface <type> <name> sync-interval <val>\n"
"  ptpdebug set interface <type> <name> delay-req-interval <val>\n"
"  ptpdebug set interface <type> <name> announce-timeout <val>\n"
"  ptpdebug set interface <type> <name> delay-mechanism [auto|end-to-end|peer-to-peer]\n"
"  ptpdebug set interface <type> <name> acceptable-master-table [enable|disable]\n"
"\n"
#endif
const char help_message[] =
"Usage:\n"
"  ptpdebug show clock [parent]\n"
"  ptpdebug show config\n"
"  ptpdebug show status \n"
"  ptpdebug show corrections [detail]\n"
"  ptpdebug show counters\n" 
"  ptpdebug show acceptable-master [log]\n"
"  ptpdebug show forced-master [log]\n"
"  ptpdebug show time-property\n"
"  ptpdebug show interface <type> <name> config \n"
"  ptpdebug show interface <type> <name> status \n"
"  ptpdebug show interface <type> <name> counters\n"
"  ptpdebug show interface <type> <name> foreign-masters\n"
"\n"
"  ptpdebug command clear-amt \n"
"  ptpdebug command clear forced-master log\n"
"  ptpdebug command clear-counters\n"
"  ptpdebug command interface <type> <name> clear-counters\n" 
"  ptpdebug command logging enable\n"
"  ptpdebug command logging disable\n"
"\n"
"";

typedef struct {
    int n_commands;
    PtpCommand *commands;
    int n_iftypes;
    PtpIfType *iftypes;
    int n_glbl_cfg_options;
    PtpCommandOption *glbl_cfg_options;
    int n_glbl_show_options;
    PtpCommandOption *glbl_show_options;
    int n_glbl_command_options;
    PtpCommandOption *glbl_command_options;
    int n_if_cfg_options;
    PtpCommandOption *if_cfg_options;
    int n_if_show_options;
    PtpCommandOption *if_show_options;
    int n_if_command_options;
    PtpCommandOption *if_command_options;
} PtpiCliElements;

PtpCommand commands[] = {
    {"set"},
    {"show"},
    {"command"},
};

PtpIfType iftypes[] = {
    {"ethernet"},
    {"bond"},
    {"vlan"},
    {"vrf"},
};

PtpCommandOption global_config_options[] = { // ptp config
    {"ptp", true, false, PTP_ARG_TYPE_BOOL },     //  [enable|disable]
    {"domain", true, false, PTP_ARG_TYPE_UINT8_VALUE},    //  domain <val> 
    {"priority1", true, false, PTP_ARG_TYPE_UINT8_VALUE}, //  priority1 <val> 
    {"priority2", true, false, PTP_ARG_TYPE_UINT8_VALUE}, //  priority2 <val> 
    {"offset-min-threshold", true, false, PTP_ARG_TYPE_VALUE},//  offset-threshold <val> 
    {"offset-max-threshold", true, false, PTP_ARG_TYPE_VALUE},//  offset-threshold <val> 
    {"mpd-threshold", true, false, PTP_ARG_TYPE_VALUE},   //  mpd-threshold <val> 
    {"acceptable-master", true, false, PTP_ARG_TYPE_CLKID},   //  acceptable-master <val> 
    {"message-mode", true, false, PTP_ARG_TYPE_MODE}, //  message-mode [multicast|unicast|mixed] 
    {"slave-only", true, false, PTP_ARG_TYPE_BOOL}, // slave-only [enable|disable]
    {"ttl", true, false, PTP_ARG_TYPE_VALUE}, //  ttl <val> 
    {"dscp-event", true, false, PTP_ARG_TYPE_DSCP_VALUE}, //  dscp-event <val> 
    {"dscp-general", true, false, PTP_ARG_TYPE_DSCP_VALUE}, //  dscp-general <val> 
};

PtpCommandOption interface_config_options[] = { // ptp config interface <type> <name>
    {"ptp", true, false, PTP_ARG_TYPE_BOOL}, //  ptp [enable|disable]
    {"ipv6", true, false, PTP_ARG_TYPE_BOOL},   //  ipv6 [enable|disable] 
    {"forced-master", true, false, PTP_ARG_TYPE_BOOL}, //   forced-master [enable|disable] 
    {"announce-interval", true, false, PTP_ARG_TYPE_INTVL}, //  announce-interval <val> 
    {"sync-interval", true, false, PTP_ARG_TYPE_INTVL}, //  sync-interval <val> 
    {"delay-req-interval", true, false, PTP_ARG_TYPE_INTVL}, //  delay-reqsync-interval <val> 
    {"announce-timeout", true, false, PTP_ARG_TYPE_UINT8_VALUE}, //   announce-timeout <val> 
    {"delay-mechanism", true, false, PTP_ARG_TYPE_DM}, //  delay-mechanism [end-to-end|peer-to-peer] 
    {"acceptable-master-table", true, false, PTP_ARG_TYPE_BOOL}, //  acceptable-master-table [enable|disable] 

};

PtpCommandOption global_show_options[] = {
    {"clock", true, true, PTP_ARG_TYPE_PARENT},   // ptp show clock [parent]
    {"config", false, false, 0},  // ptp show config 
    {"status", false, false, 0},  // ptp show status 
    {"counters", false, false, 0},   // ptp show counters 
    {"acceptable-master", true, true, PTP_ARG_TYPE_LOG}, // ptp show acceptable-master [log]
    {"forced-master", true, true, PTP_ARG_TYPE_LOG}, //ptp show forced-master [log]
    {"time-property", false, true, 0}, // ptp show time-property
};

PtpCommandOption interface_show_options[] = {
    {"config", false, false, 0},   // ptp show interface <type> <name> config 
    {"status", false, false, 0},   // ptp show interface <type> <name> status 
    {"counters", false, false, 0}, // ptp show interface <type> <name> counters
    {"foreign-masters", false, false, 0}, // ptp show interface <type> <name> foreign-masters
};

PtpCommandOption global_command_options[] = {
    {"clear-amt-log", true, true, 0},   // ptp clear-amt [log]
    {"clear-fm-log", true, false, 0},   // ptp clear-forced-masters log
    {"clear-counters", false, false, 0},   // ptp clear-counters 
    {"logging", true, false, PTP_ARG_TYPE_BOOL },     //  ptp logging [enable|disable]
};

PtpCommandOption interface_command_options[] = {
    {"clear-counters", false, false, 0},   // ptp clear interface <type> <name> counters 
};

PtpiCliElements elements = {
     3, commands,
     4, iftypes,
     13, global_config_options,
     7, global_show_options,
     4, global_command_options,
     9, interface_config_options,
     4, interface_show_options,
     1, interface_command_options,
};

/*
 * Tokens object
 */

typedef struct Tokens {
    int argc;
    char **argv;
    int i;
    char *current;
} Tokens;

Tokens tokens_new(int argc, char **argv) 
{
    Tokens ts = {argc, argv, 0, argv[0]};
    return ts;
}

Tokens* tokens_move(Tokens *ts) 
{
    if (ts->i < ts->argc) {
        ts->current = ts->argv[++ts->i];
    }
    if (ts->i == ts->argc) {
        ts->current = NULL;
    }
    return ts;
}

int parse_argcmd(Tokens *ts, PtpiCliElements *elements, PtpParsedArgs *args) 
{
    int i;
    PtpCommand *command;
    PtpCommand *commands = elements->commands;

    for (i=0; i < elements->n_commands; i++) {
        command = &commands[i];
	if (args->debug_enable)
            fprintf(stdout, "%s: commandtbl[%d]:%s tokencmd:%s\n", __FUNCTION__, i,
			command->name, ts->current);
        if (!strcmp(command->name, ts->current)){
	    args->cmd_type = i + 1;
            tokens_move(ts);
            return 0;
        }
    }
    return -1;
}

static int
parse_and_validate_arg_options(char *current, PtpCommandOption *option,
		PtpParsedArgs *args)
{
    switch (option->arg_type) {
    	case PTP_ARG_TYPE_BOOL:
			if (!strcmp("enable", current)) {
				args->opt_bool_argument = true;
			} else if (!strcmp("disable", current)) {
				args->opt_bool_argument = false;
			} else {
				goto invalid_option;
			}	
			break;
   		 case PTP_ARG_TYPE_MODE:
			if (!strcmp("unicast", current)) {
				args->opt_int_argument = 1;
			} else if (!strcmp("multicast", current)) {
				args->opt_int_argument = 2;
			} else if (!strcmp("mixed", current)) {
				args->opt_int_argument = 3;
			} else {
				goto invalid_option;
			}	
			break;
   		 case PTP_ARG_TYPE_PARENT:
			if (strcmp("parent", current)) {
				goto invalid_option;
			}
			break;
   		 case PTP_ARG_TYPE_LOG:
			if (strcmp("log", current)) {
				goto invalid_option;
			}
			break;
   		 case PTP_ARG_TYPE_DETAIL:
			if (strcmp("detail", current)) {
				goto invalid_option;
			}
			break;
   		 case PTP_ARG_TYPE_CLKID:
			args->opt_str_argument = current;
			break;
   		 case PTP_ARG_TYPE_VALUE:
			args->opt_int_argument = atoi(current);
			break; 
   		 case PTP_ARG_TYPE_INTVL:
			args->opt_int_argument = atoi(current);
			if (((args->opt_int_argument) > 8) || 
			    ((args->opt_int_argument) < -8)) {
				goto invalid_option;
			}
			break; 
   		 case PTP_ARG_TYPE_DM:
			if (!strcmp("auto", current)) {
				args->opt_int_argument = DM_AUTO;
			} else if (!strcmp("end-to-end", current)) {
				args->opt_int_argument = DM_E2E;
			} else if (!strcmp("peer-to-peer", current)) {
				args->opt_int_argument = DM_P2P;
			} else {
				goto invalid_option;
			}
			break; 
   		 default:
			break;
   		 }	
   		 return 0;
invalid_option: 
    fprintf(stderr, "%s - is invalid option\n %s", 
        current, help_message); 
    return -1;
}

int parse_argif(Tokens *ts, PtpiCliElements *elements, PtpParsedArgs *args) 
{
    int i;
    PtpIfType *iftype;
    PtpIfType *iftypes = elements->iftypes;

    for (i=0; i < elements->n_iftypes; i++) {
        iftype = &iftypes[i];
	if (args->debug_enable)
	    fprintf(stdout, "parse_argif i:%d cname:%s ts->c:%s\n", i,
			iftype->name, ts->current);
        if (!strcmp(iftype->name, ts->current)){
	    args->interface_type = i + 1;

	    // Move to argument
            tokens_move(ts);

            if (ts->current == NULL) {
                fprintf(stderr, 
 		    "IF:%s requires argument\n", iftype->name);
                return -1;
            }

	    args->type_argument = ts->current;

            tokens_move(ts);
	    return 0;
        }
    }
    return -1;
}

int parse_arg_ifopt(Tokens *ts, PtpiCliElements *elements, PtpParsedArgs *args) 
{
    int i, n_options = 0;
    PtpCommandOption *option;
    PtpCommandOption *options;

    switch (args->cmd_type) {
	case PTP_ARG_CMD_TYPE_SET:
	    options = elements->if_cfg_options;
	    n_options = elements->n_if_cfg_options;
	    break;
	case PTP_ARG_CMD_TYPE_GET:
	    options = elements->if_show_options;
	    n_options = elements->n_if_show_options;
	    break;
	case PTP_ARG_CMD_TYPE_COMMAND:
	    options = elements->if_command_options;
	    n_options = elements->n_if_command_options;
	    break;
    }
    args->if_cfg_opt = 0;
    args->if_show_opt = 0;
    args->if_command_opt = 0;
    for (i=0; i < n_options; i++) {
        option = &options[i]; 
		
		if (args->debug_enable) 
			fprintf(stdout, "argifopt: optionnum:%d option:%s ts->c:%s\n", 
						i, option->name, ts->current);

        if (!strcmp(option->name, ts->current)){ 
				
			switch (args->cmd_type) { 
				case PTP_ARG_CMD_TYPE_SET:
    	            args->if_cfg_opt = i + 1;
		    		break; 
				case PTP_ARG_CMD_TYPE_GET:
    	            args->if_show_opt = i + 1;
		    		break; 
				case PTP_ARG_CMD_TYPE_COMMAND:
    	            args->if_command_opt = i + 1;
		    		break;
            }

		    // Move to argument
            tokens_move(ts);
            args->opt_takes_arg = option->takes_arg;
            args->opt_arg_optional = option->arg_optional;
            args->opt_arg_present = false;
	    
		    if (ts->current == NULL) { 
				if (!option->takes_arg || (option->takes_arg && option->arg_optional)) { 
			    	return 0; 
		    	} 
		    	fprintf(stderr, 
				    "%s IF option:%s requires argument\n", 
				    (args->cmd_type == PTP_ARG_CMD_TYPE_SET) ?  "config" : 
				   (args->cmd_type == PTP_ARG_CMD_TYPE_GET) ?  
				   "show" : "command", option->name); 
			    return -1;
            }

            args->opt_arg_present = true;
		    args->arg_type = option->arg_type; 
	    
		    if (parse_and_validate_arg_options(ts->current, option, args)) {
			    return -1;
		    }
		    tokens_move(ts);
            return 0;
        }
    }
    return -1;
}

int parse_arg_glblopt(Tokens *ts, PtpiCliElements *elements, PtpParsedArgs *args) 
{
    int i, n_options = 0;
    PtpCommandOption *option;
    PtpCommandOption *options;

    switch (args->cmd_type) {
	case PTP_ARG_CMD_TYPE_SET:
	    options = elements->glbl_cfg_options;
	    n_options = elements->n_glbl_cfg_options;
	    break;
	case PTP_ARG_CMD_TYPE_GET:
	    options = elements->glbl_show_options;
	    n_options = elements->n_glbl_show_options;
	    break;
	case PTP_ARG_CMD_TYPE_COMMAND:
	    options = elements->glbl_command_options;
	    n_options = elements->n_glbl_command_options;
	    break;
    }
    args->glbl_cfg_opt = 0;
    args->glbl_show_opt = 0;
    args->glbl_command_opt = 0;
    for (i=0; i < n_options; i++) {
        option = &options[i];
		if (args->debug_enable)
			fprintf(stdout, "%s: optiontbl[%d]:%s tokenoption:%s\n", __FUNCTION__,i,
				option->name, ts->current); 
	
		if (!strcmp(option->name, ts->current)){
	    	switch (args->cmd_type) {
	        	case PTP_ARG_CMD_TYPE_SET:
   	            	args->glbl_cfg_opt = i + 1;
				    break;
		        case PTP_ARG_CMD_TYPE_GET:
 		            args->glbl_show_opt = i + 1;
				    break;
		        case PTP_ARG_CMD_TYPE_COMMAND:
  		            args->glbl_command_opt = i + 1;
				    break;
   		     } 
			
			// Move to argument
			tokens_move(ts); 
			
			args->opt_takes_arg = option->takes_arg; 
			args->opt_arg_optional = option->arg_optional; 
			args->opt_arg_present = false; 
			
			if (ts->current == NULL) { 
				if (!option->takes_arg || (option->takes_arg && option->arg_optional)) { 
					return 0; 
				} 
				
				fprintf(stderr, "%s glbl option:%s requires argument\n", 
						(args->cmd_type == PTP_ARG_CMD_TYPE_SET) ?  "set" : 
						(args->cmd_type == PTP_ARG_CMD_TYPE_GET) ?  
						"get" : "command", option->name); 
				return -1; 
			} 
			args->opt_arg_present = true; 
			args->arg_type = option->arg_type; 
			if (parse_and_validate_arg_options(ts->current, option, args)) { 
				return -1; 
			} 
			tokens_move(ts);
            return 0;
        }
    }
    return -1;
}

int parse_args(Tokens *ts, PtpiCliElements *elements, PtpParsedArgs *args) 
{
    int ret;

    tokens_move(ts);
    while (ts->current != NULL) {
	if (!strcmp(ts->current, "-d")) {
            args->debug_enable = true;	
	    tokens_move(ts);
	    continue;
	} else if (!strcmp(ts->current, "-h")) {
	    fprintf(stdout, "%s", help_message);
	}
        ret = parse_argcmd(ts, elements, args);
        if (ret) {
		fprintf(stdout, "%s", help_message);
		return ret;
	}
	if (ts->current == NULL) {
		fprintf(stdout, "%s", help_message);
		return -1;
	}
	if (!strcmp(ts->current, "interface")) {
	    if (args->debug_enable)
		fprintf(stdout, "Interface command\n");
            args->global = false;
	    tokens_move(ts);

	    // Check interface <type> <name>
            ret = parse_argif(ts, elements, args);
            if (ret) return ret;

	    if (ts->current == NULL) return 0;
	    
	    // Parse if command options
            ret = parse_arg_ifopt(ts, elements, args);
            if (ret) return ret;
        } else {
            args->global = true;

	    // Parse global command options
            ret = parse_arg_glblopt(ts, elements, args);
            if (ret) return ret;
        }
    }
    return 0;
}

/*
 * Main command, arguments and options parse function
 */

PtpParsedArgs ptp_cmd_parser(int argc, char *argv[]) 
{
    PtpParsedArgs args;
    Tokens ts;

    memset(&args, 0, sizeof(PtpParsedArgs));
    args.help_message = help_message;

    ts = tokens_new(argc, argv);
    if (parse_args(&ts, &elements, &args))
        exit(EXIT_FAILURE);
    return args;
}
