/**
 * @file ts_monitor.h
 * Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * @note Copyright (C) 2021 Saravanan Arun <sarun@nvidia.com>
 * @note SPDX-License-Identifier: GPL-2.0+
 */
#ifndef HAVE_TS_MONITOR_H
#define HAVE_TS_MONITOR_H

#include "config.h"
#include "port.h"
#include "tmv.h"

#define DEFAULT_TIME_STAMP_ENTRIES  400
#define DEFAULT_LOG_SETS 8
#define DEFAULT_LOG_ENTRIES 8

struct time_stamp_entry {
        uint16_t       sync_seqid;
        tmv_t          t1;
        tmv_t          t2;
        tmv_t          t3;
        tmv_t          t4;
        tmv_t          master_offset;
        tmv_t          path_delay;
        struct timeval sync_time;
};

struct time_stamp_data {
        uint32_t       wr_idx;
	uint32_t       offset_min_vio_log_index;
	uint32_t       offset_max_vio_log_index;
	uint32_t       mpd_vio_log_index;

	uint32_t       offset_min_violations;
	uint32_t       offset_max_violations;
	uint32_t       mpd_violations;

	uint32_t       num_ts;
	uint32_t       num_log_sets;
	uint32_t       num_log_entries;
	uint32_t       log_wait_secs;

	int32_t        offset_min_threshold;
	int32_t        offset_max_threshold;
	int32_t        mpd_threshold;

	struct timeval last_offset_min_violation;
	struct timeval last_offset_max_violation;
	struct timeval last_mpd_violation;
};

struct ts_monitor *ts_monitor_create(struct config *config, struct tsproc *tsp);

void ts_monitor_destroy(struct ts_monitor *monitor);

void ts_monitor_update(struct ts_monitor *tsm, uint16_t seqid, tmv_t t1, tmv_t t2);

#endif
